unit UMatSky;
{$DEFINE DBGMATSKY}
{diverses routines de calcul sur matrices ligne de ciel
 rsolution par mthode de Cholesky
}
interface

uses UMatriceFortran2                          ,dialogs
     ;

type
  TProcMatSkyProgress = procedure(var stop:boolean);

  PCellMatSky = ^double;

  TMatSky = class
  private
    nomM: TNomMV;
    dim: integer; //dimension
    cols: array of pointer;//pointe sur chaque colonne
    hcols: array of integer;//les hauteurs de colonnes
    Ftri: boolean; //true si la matrice est triangule
    function GetCellSky(col,pos:integer):double;
    procedure SetCellSky(col,pos:integer;Value:double);
    function GetCellStd(lig,col:integer):double;
    procedure SetCellStd(lig,col:integer;Value:double);
    function GetCellStd0(lig,col:integer):double;
  public
    constructor Create(nom:string; ncol:integer);
    constructor CreateCopy(nom:string; SRC:TMatSky; struconly:boolean=false);
    //si struconly, copie la taille des colonnes, mais pas le contenu ; remplit avec des 0
    destructor Destroy;override;
    property NbCols:integer read dim;     //la taille
    procedure MinHtCol(col,hmin:integer); //fixe la ht mini de chaque colonne
    procedure InitCiel;//rserve la mmoire de la matrice en fonction des hauteurs colonnes
    function NbElts:integer; //contenu de la matrice
    procedure ClearData;//met  0 mais garde la mmoire
    function LoadData(PF:string):string;//erreur si <> ''
    function SaveData(PF:string):string;//erreur si <> ''
    procedure CopyData(SRC:TMatSky); //copie le contenu des colonnes (complte CreateCopy/struconly)
    function TexteMat: string; //la matrice complte, sp tabs
    function TexteMat2: string; //la matrice complte, sp tabs
    function PCellSky(col,pos:integer):PCellMatSky; //renvoie un pointeur sur la cellule ; col=1.., pos=1 sur la diag
    function PCellStd(lig,col:integer):PCellMatSky; //idem, adressage rectangulaire : 1.. , 1..
    property CellSky[col,pos:integer]:double read GetCellSky write SetCellSky;//lit et crit directement, mme adressage que ci-dessus
    property CellStd[lig,col:integer]:double read GetCellStd write SetCellStd;
    property CellStd0[lig,col:integer]:double read GetCellStd0; //renvoie 0 si hors matrice
    function DebutColStd(col:integer):integer; //1re ligne non nulle de la colonne
    function InMat(lig,col:integer):boolean; //si tombe dans du "plein"
    procedure MxV(var Vres:TVecDouble;var V:TVecDouble;n:integer); //produit matrice x vecteur : Vres = Mat * V
    procedure Add(coef:double;mat:TMatSky); //matrice = matrice + mat * coef
    function CheckCompat(mat:TMatSky; raiseerr:boolean):boolean; //vrifie que les structures des 2 matrices (htcol, etc) sont identiques
    property Triangulee: boolean read Ftri; //si la matrice a "subi" Cholesky_tri

    function Cholesky_tri(n:integer; progress:TProcMatSkyProgress; TestTousLes:integer ):integer; //triangule la sous-mat 1..n
    function Cholesky_solve(BX:PTblD;n:integer; progress:TProcMatSkyProgress; TestTousLes:integer ):integer; //rsoud mat (n) * X = B ; BX = B puis X
  end;

implementation

uses Forms, SysUtils, Math;

const
  k_erpos = 'position hors de la colonne';
  k_ercol = 'colonne hors limites';
  k_erlig = 'ligne hors limites';
  k_erhc  = 'MinHtCol interdit aprs InitCiel';

constructor TMatSky.Create(nom:string; ncol:integer);
begin
  inherited Create;
  nomM:=nom;
  dim:=ncol;
  SetLength(hcols,ncol);
end;

constructor TMatSky.CreateCopy(nom:string; SRC:TMatSky; struconly:boolean=false);
var
i,k: integer;
begin
  Create(nom,SRC.dim);
  SetLength(cols,dim);
  for i:=0 to dim-1 do begin
    k:=SRC.hcols[i];//ht col i
    hcols[i]:=k;
    GetMem(cols[i],k*sizeof(double));
    if struconly
      then FillChar(cols[i]^,k*sizeof(double),0)      //vide
      else Move(SRC.cols[i]^,cols[i]^,k*sizeof(double));  //copie
  end;
end;

destructor TMatSky.Destroy;
var
i: integer;
begin
  if Assigned(cols) then begin
    for i:=0 to dim-1 do FreeMem(cols[i],hcols[i]*sizeof(double));
    Finalize(cols);
  end;
  Finalize(hcols);
  inherited;
end;

procedure TMatSky.CopyData(SRC:TMatSky);
var
i: integer;
begin
  CheckCompat(SRC,true); //vrifie que la structure est la mme, erreur sinon
  for i:=0 to dim-1 do Move(SRC.cols[i]^,cols[i]^,hcols[i]*sizeof(double));  //copie
end;

procedure TMatSky.MinHtCol(col,hmin:integer);
//dfinit une hauteur mini pour la colonne
begin
  if Assigned(cols) then raise Exception.Create(k_erhc);
  hcols[col-1]:=Max( hcols[col-1] , hmin );
end;

procedure TMatSky.InitCiel;
//rserve la mmoire
var
i,k: integer;
begin
  SetLength(cols,dim);//seulement ici
  for i:=0 to dim-1 do begin
    k:=hcols[i];//ht col i
    if k=0 then raise Exception.Create(nomM+' : hcol=0 en '+IntToStr(i+1));
    GetMem(cols[i],k*sizeof(double));
  end;
  ClearData;
end;

function TMatSky.NbElts:integer;
var
i: integer;
begin
  result:=0;
  for i:=0 to dim-1 do inc(result,hcols[i]);
end;

procedure TMatSky.ClearData;
//rserve la mmoire
var
i: integer;
begin
  Ftri:=false;
  for i:=0 to dim-1 do FillChar(cols[i]^,hcols[i]*sizeof(double),0);
end;

function TMatSky.SaveData(PF:string):string;//erreur si <> ''
//ne range que les cellules de la matrice
var
F:file;
i: integer;
begin
  result:='';
  try
    AssignFile(F,PF);
    Rewrite(F,1);
    BlockWrite(F,Ftri,sizeof(boolean));
    for i:=1 to dim do begin
      BlockWrite(F,cols[i-1]^,hcols[i-1]*sizeof(double));
      Application.ProcessMessages;
    end;
    CloseFile(F);
  except
    on E:exception do result:=E.Message;
  end;
end;

function TMatSky.LoadData(PF:string):string;//erreur si <> ''
var
F:file;
i: integer;
begin
  result:='';
  try
    AssignFile(F,PF);
    Reset(F,1);
    BlockRead(F,Ftri,sizeof(boolean));
    for i:=1 to dim do begin
      BlockRead(F,cols[i-1]^,hcols[i-1]*sizeof(double));
      Application.ProcessMessages;
    end;
    CloseFile(F);
  except
    on E:exception do result:=E.Message;
  end;
end;

function TMatSky.TexteMat: string; //la matrice complte, sp tabs
var
li,co: integer;
d: double;
S: string;
begin
  for li:=1 to dim do begin
    S:='';
    for co:=1 to dim do begin
      if co<li then S:=S+#9
      else begin
        if InMat(li,co) then d:=CellStd[li,co] else d:=0;
        S:=S+Format('%.3e',[d])+#9;
      end;
    end;
    result:=result+S+#13#10;
  end;
end;

function TMatSky.TexteMat2: string; //la matrice complte, sp tabs, vide hors colonne
var
li,co: integer;
d: double;
S: string;
begin
  for li:=1 to dim do begin
    S:='';
    for co:=1 to dim do begin
      if co<li then S:=S+#9
      else begin
        if InMat(li,co) then S:=S+Format('%.3e',[CellStd[li,co]]);
        S:=S+#9;
      end;
    end;
    result:=result+S+#13#10;
  end;
end;

function TMatSky.GetCellSky(col,pos:integer):double;
begin
{$IFDEF DBGMATSKY}
  if (col<1) or (col>dim) then raise Exception.Create(k_ercol); if pos>hcols[col-1] then raise Exception.Create(k_erpos);
{$ENDIF}
  result:=PCellMatSky(pchar(cols[col-1])+(pos-1)*sizeof(double))^;
end;

procedure TMatSky.SetCellSky(col,pos:integer;Value:double);
begin
{$IFDEF DBGMATSKY}
  if (col<1) or (col>dim) then raise Exception.Create(k_ercol); if pos>hcols[col-1] then raise Exception.Create(k_erpos);
{$ENDIF}
  PCellMatSky(pchar(cols[col-1])+(pos-1)*sizeof(double))^:=Value;
end;

function TMatSky.GetCellStd(lig,col:integer):double;
begin
  if lig>col then begin //on permute lig et col (les 2 sont cods pour gagner un peu de temps)
{$IFDEF DBGMATSKY}
    if (lig<1) or (lig>dim) then raise Exception.Create(k_erlig); if lig-col+1>hcols[lig-1] then raise Exception.Create(k_ercol);
{$ENDIF}
    result:=PCellMatSky(pchar(cols[lig-1])+(lig-col)*sizeof(double))^;
  end else begin //sens normal
{$IFDEF DBGMATSKY}
    if (col<1) or (col>dim) then raise Exception.Create(k_ercol); if col-lig+1>hcols[col-1] then raise Exception.Create(k_erlig);
{$ENDIF}
    result:=PCellMatSky(pchar(cols[col-1])+(col-lig)*sizeof(double))^;
  end;
end;

procedure TMatSky.SetCellStd(lig,col:integer;Value:double);
begin
  if lig>col then begin //on permute lig et col (les 2 sont cods pour gagner un peu de temps)
{$IFDEF DBGMATSKY}
    if (lig<1) or (lig>dim) then raise Exception.Create(k_erlig); if lig-col+1>hcols[lig-1] then raise Exception.Create(k_ercol);
{$ENDIF}
    PCellMatSky(pchar(cols[lig-1])+(lig-col)*sizeof(double))^:=Value;
  end else begin //sens normal
{$IFDEF DBGMATSKY}
    if (col<1) or (col>dim) then raise Exception.Create(k_ercol); if col-lig+1>hcols[col-1] then raise Exception.Create(k_erlig);
{$ENDIF}
    PCellMatSky(pchar(cols[col-1])+(col-lig)*sizeof(double))^:=Value;
  end;
end;

function TMatSky.GetCellStd0(lig,col:integer):double;
begin
  if lig>col then begin //on permute lig et col (les 2 sont cods pour gagner un peu de temps)
    if (lig<1) or (lig>dim) or (lig-col+1>hcols[lig-1])
     then result:=0
     else result:=PCellMatSky(pchar(cols[lig-1])+(lig-col)*sizeof(double))^;
  end else begin //sens normal
    if (col<1) or (col>dim) or (col-lig+1>hcols[col-1])
     then result:=0
     else result:=PCellMatSky(pchar(cols[col-1])+(col-lig)*sizeof(double))^;
  end;
end;

function TMatSky.PCellSky(col,pos:integer):PCellMatSky;
begin
{$IFDEF DBGMATSKY}
  if (col<1) or (col>dim) then raise Exception.Create(k_ercol); if pos>hcols[col-1] then raise Exception.Create(k_erpos);
{$ENDIF}
  result:=PCellMatSky(pchar(cols[col-1])+(pos-1)*sizeof(double));
end;

function TMatSky.PCellStd(lig,col:integer):PCellMatSky;
begin
  if lig>col then begin //on permute lig et col (les 2 sont cods pour gagner un peu de temps)
{$IFDEF DBGMATSKY}
    if (lig<1) or (lig>dim) then raise Exception.Create(k_erlig); if lig-col+1>hcols[lig-1] then raise Exception.Create(k_ercol);
{$ENDIF}
    result:=PCellMatSky(pchar(cols[lig-1])+(lig-col)*sizeof(double));
  end else begin //sens normal
{$IFDEF DBGMATSKY}
    if (col<1) or (col>dim) then raise Exception.Create(k_ercol); if col-lig+1>hcols[col-1] then raise Exception.Create(k_erlig);
{$ENDIF}
    result:=PCellMatSky(pchar(cols[col-1])+(col-lig)*sizeof(double));
  end;
end;

function TMatSky.DebutColStd(col:integer):integer;
//renvoit la ligne de dbut haut de colonne en format (lig,col)
begin
{$IFDEF DBGMATSKY}
    if (col<1) or (col>dim) then raise Exception.Create(k_ercol);
{$ENDIF}
  result := col - ( hcols[col-1] - 1 );
end;

function TMatSky.InMat(lig,col:integer):boolean;
begin
  result:=false;
  if lig>col then begin //on permute lig et col (les 2 sont cods pour gagner un peu de temps)
    if (lig<1) or (lig>dim) or (lig-col+1>hcols[lig-1]) then exit;
  end else begin //sens normal
    if (col<1) or (col>dim) or (col-lig+1>hcols[col-1]) then exit;
  end;
  result:=true;
end;

procedure TMatSky.MxV(var Vres:TVecDouble;var V:TVecDouble;n:integer);
var
  li,k: integer;
  d: double;
begin
  for li:=1 to n do begin
    d:=0;
    for k:=1 to n do d:=d+CellStd0[li,k] * PCellRelD(V,k)^;
    PCellRelD(Vres,li)^:=d;
  end;
end;

procedure TMatSky.Add(coef:double;mat:TMatSky);
var
i,j,k: integer;
begin                 
//vrifier compatibilit
  CheckCompat(mat,true);
//ajouter
  for i:=0 to dim-1 do begin
    k:=hcols[i];//ht col i
    for j:=1 to k do PTblD(cols[i])^[j] := PTblD(cols[i])^[j] + coef * PTblD(mat.cols[i])^[j] ;
  end;
end;

function TMatSky.CheckCompat(mat:TMatSky; raiseerr:boolean):boolean; //vrifie que les structures des 2 matrices (htcol, etc) sont identiques
var
i: integer;
begin
  result:= false;
  try
    if (Ftri <> mat.Ftri) then exit;
    if (length(hcols) <> length(mat.hcols)) then exit;
    for i:=0 to dim-1 do if hcols[i] <> mat.hcols[i] then exit;
    result:=true;
  finally
    if (not result) and raiseerr then raise Exception.Create(nomM + ' , ' + mat.nomM + ' : structures incompatibles');
  end;
end;

function TMatSky.Cholesky_tri(n:integer; progress:TProcMatSkyProgress; TestTousLes:integer ):integer;
//resultat 0 si ok, n de la composante en erreur sinon, -1 si arrt externe
//self est symtrique dfinie positive
//n est la taille de la sous-matrice  trianguler (en partant de 1)
//progress est une fonction de rappel informant de la progression et testant la demande d'arrt
//TestTousLes : nombre de boucles ppales entre 2 tests
//en retour, self contient le rsultat dans (1:n,1:n)
var
dia,dia1,lig,col,debl,debl1: integer;
Adia,d: double;
pcellDi,pcell1: PCellMatSky;
stop: boolean;
begin
  if Ftri then raise Exception.Create(nomM+' : la matrice est dj triangule');
  result:=0;
  try
    for dia:=1 to n do if CellSky[dia,1]<=0 then begin
      result:=dia;
      exit;
    end;
    for dia:=1 to n do begin
      if Assigned(progress) and (TestTousLes>0) and ((dia mod TestTousLes)=0) then begin
        stop:=false;
        progress(stop);
        if stop then begin
          result:=-1;
          exit;
        end;
      end;
      dia1:=dia-1;
      pcellDi:=PCellStd(dia,dia); //l'lment diagonal
      Adia:=PcellDi^;
      debl:=DebutColStd(dia);
      for lig:=debl to dia1 do begin  //normalisation (on descend la ligne dans la colonne dia)
        Adia:=Adia - SQR( CellStd[lig,dia] );
      end;
      if Adia<0 then raise EMathError.Create('triangulation de Cholesky : valeur ngative sur la diagonale');
      pcellDi^:=SQRT(Adia);
      for lig:=dia+1 to n do if InMat(dia,lig) then begin
        pcell1:=PCellStd(dia,lig);
        d:=pcell1^;
        debl1:=Max( debl , DebutColStd(lig) );//on dmarre du + bas car au dessus le produit fait 0
        for col:=debl1 to dia1 do begin//if (PkMatInMat(A,col,lig) and PkMatInMat(A,dia1,col)) then begin
          d:=d - CellStd[col,dia] * CellStd[col,lig];
        end;
        Adia:=PcellDi^;
        if ABS(Adia)<1E-100 then raise EMathError.Create('triangulation de Cholesky : division par zro');
        pcell1^:= d / Adia;
      end;
    end;
    Ftri:=true;
  except
    on exception do result:=dia;
  end;
end;

function TMatSky.Cholesky_solve(BX:PTblD;n:integer; progress:TProcMatSkyProgress; TestTousLes:integer ):integer;
//resultat 0 si ok, n de la composante en erreur sinon
//BX est un vecteur B tel que A * X = B (1..)
//A(=self) est une matrice triangulaire suprieure (U) skyline, obtenue par dcomposition de Cholesky (Cholesky_tri),
// telle que  trans(U) * U = A
//n est la taille du systme  rsoudre
//en retour, BX contient les vecteurs X ; A est inchange
var
dia,col: integer;
d: double;
stop: boolean;
begin
  if not Ftri then raise Exception.Create(nomM+' : la matrice n''est pas triangule');
  TestTousLes:=TestTousLes*2;
  result:=0;
  try
//on rsoud   trans(U) * Z = B, Z[i] remplace B[i], ce qui est + efficace pour l'algorithme
//les appels  CellStd considrent que la matrice est symtrique, ce qui n'est plus le cas ici, car A est est une tri (U)
//ils permettent d'obtenir "gratuitement" la transpose,  condition de n'exploiter que la partie comprise entre
//la 1re col et la diagonale
    for dia:=1 to n do begin
////
      if Assigned(progress) and (TestTousLes>0) and ((dia mod TestTousLes)=0) then begin
        stop:=false;
        progress(stop);
        if stop then begin
          result:=-1;
          exit;
        end;
      end;
////
      d:=0;
      for col:=1 to dia-1 do if InMat(dia,col) then d:=d + CellStd[dia,col] * BX[col] ;
      BX[dia]:=(BX[dia]-d)/CellStd[dia,dia];
    end;//fait
//on rsoud   U * X = Z en remontant et en n'utilisant que le U de la matrice
    for dia:=n downto 1 do begin
////
      if Assigned(progress) and (TestTousLes>0) and ((dia mod TestTousLes)=0) then begin
        stop:=false;
        progress(stop);
        if stop then begin
          result:=-1;
          exit;
        end;
      end;
////
      d:=0;
      for col:=dia+1 to n do if InMat(dia,col) then d:=d + CellStd[dia,col] * BX[col] ;
      BX[dia]:=(BX[dia]-d)/CellStd[dia,dia];
    end;//fait
  except
    on E:exception do
      if E is EAbort
      then raise
      else result:=dia;
  end;
end;

end.
